#!/usr/bin/env python
#
# Copyright 2007 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#



"""
In-memory persistent stub for the Python datastore API. Gets, queries,
and searches are implemented as in-memory scans over all entities.

Stores entities across sessions as pickled proto bufs in a single file. On
startup, all entities are read from the file and loaded into memory. On
every `Put()`, the file is wiped and all entities are written from scratch.
Clients can also manually `Read()` and `Write()` the file themselves.
"""

import collections
import logging
import os
import struct
import tempfile
import threading
import weakref

import six
from six.moves import filter
from six.moves import map
import six.moves.cPickle as pickle

from google.appengine.api import apiproxy_stub
from google.appengine.api import cmp_compat
from google.appengine.api import datastore
from google.appengine.api import datastore_types
from google.appengine.datastore import datastore_pb
from google.appengine.datastore import datastore_stub_util
from google.appengine.runtime import apiproxy_errors
from google.protobuf import message
from google.appengine.datastore import entity_bytes_pb2 as entity_pb2
















datastore_pb.Query.__hash__ = lambda self: hash(self.SerializeToString())

logger = logging.getLogger('google.appengine.api.stubs.datastore')

if six.PY3:
  long = int


def _FinalElement(key):
  """Return final element of a key's path."""
  return key.path.element[-1]


class _StoredEntity(object):
  """Simple wrapper around an entity (and its metadata) stored by the stub.

  Public properties:
    record: The original `EntityRecord` that was stored by the stub.
    encoded_protobuf: Encoded binary representation of entity protobuf,
        including a special property that holds the pickled metadata object.
  """

  def __init__(self, record):
    """Creates a `_StoredEntity` object and store an entity.

    Args:
      record: The `EntityRecord` to store.
    """
    self.record = record






    entity = datastore_stub_util._ToStorageEntity(record)
    self.encoded_protobuf = entity.SerializeToString()


class KindPseudoKind(object):
  """Pseudo-kind for schema queries.

  Provides a Query method to perform the actual query.

  Public properties:
    name: The pseudo-kind name
  """
  name = '__kind__'

  def Query(self, query, filters, orders):
    """Perform a query on this pseudo-kind.

    Args:
      query: The original `datastore_pb.Query`.
      filters: The filters from query.
      orders: The orders from query.

    Returns:
      `(results, remaining_filters, remaining_orders)`.
      `results` is a list of `entity_pb2.EntityProto`.
      `remaining_filters` and `remaining_orders` are the filters and orders that
      should be applied in memory.
    """
    kind_range = datastore_stub_util.ParseKindQuery(query, filters, orders)
    app_namespace_str = datastore_types.EncodeAppIdNamespace(
        query.app, query.name_space)
    kinds = []


    for app_namespace, kind in self._stub._GetAllEntities():
      if app_namespace != app_namespace_str: continue
      if not kind_range.Contains(kind): continue
      kinds.append(
          datastore.Entity(
              self.name, name=kind, _app=query.app,
              namespace=query.name_space)._ToPb())

    return (kinds, [], [])


class PropertyPseudoKind(object):
  """Pseudo-kind for schema queries.

  Provides a Query method to perform the actual query.

  Public properties:
    name: The pseudo-kind name.
  """
  name = '__property__'

  def Query(self, query, filters, orders):
    """Perform a query on this pseudo-kind.

    Args:
      query: The original `datastore_pb.Query`.
      filters: The filters from query.
      orders: The orders from query.

    Returns:
      `(results, remaining_filters, remaining_orders)`.
      `results` is a list of `entity_pb2.EntityProto`.
      `remaining_filters` and `remaining_orders` are the filters and orders that
      should be applied in memory.
    """
    property_range = datastore_stub_util.ParsePropertyQuery(query, filters,
                                                            orders)
    keys_only = query.keys_only
    app_namespace_str = datastore_types.EncodeAppIdNamespace(
        query.app, query.name_space)

    properties = []
    if keys_only:
      usekey = '__property__keys'
    else:
      usekey = '__property__'

    entities = self._stub._GetAllEntities()
    for app_namespace, kind in entities:
      if app_namespace != app_namespace_str: continue

      app_kind = (app_namespace_str, kind)



      (start_cmp, end_cmp) = property_range.MapExtremes(
          lambda extreme, inclusive, is_end: cmp_compat.cmp(kind, extreme[0]))
      if not((start_cmp is None or start_cmp >= 0) and
             (end_cmp is None or end_cmp <= 0)):
        continue


      kind_properties = self._stub._GetSchemaCache(app_kind, usekey)
      if not kind_properties:
        kind_properties = []
        kind_key = datastore_types.Key.from_path(
            KindPseudoKind.name,
            kind,
            _app=query.app,
            namespace=query.name_space)

        props = collections.defaultdict(set)



        for entity in entities[app_kind].values():
          for prop in entity.record.entity.property:
            prop_name = prop.name

            if (prop_name in
                datastore_stub_util.GetInvisibleSpecialPropertyNames()):
              continue
            value_pb = prop.value
            props[prop_name].add(datastore_types.GetPropertyValueTag(value_pb))


        for prop in sorted(props):
          property_e = datastore.Entity(
              self.name,
              name=prop,
              parent=kind_key,
              _app=query.app,
              namespace=query.name_space)

          if not keys_only and props[prop]:
            property_e['property_representation'] = [
                datastore_stub_util._PROPERTY_TYPE_NAMES[tag]
                for tag in sorted(props[prop])]

          kind_properties.append(property_e._ToPb())

        self._stub._SetSchemaCache(app_kind, usekey, kind_properties)


      def InQuery(property_e):
        return property_range.Contains((kind,
                                        _FinalElement(property_e.key).name))

      properties += list(filter(InQuery, kind_properties))

    return (properties, [], [])


class NamespacePseudoKind(object):
  """Pseudo-kind for namespace queries.

  Provides a Query method to perform the actual query.

  Public properties:
    name: The pseudo-kind name.
  """
  name = '__namespace__'

  def Query(self, query, filters, orders):
    """Perform a query on this pseudo-kind.

    Args:
      query: The original `datastore_pb.Query`.
      filters: The filters from query.
      orders: The orders from query.

    Returns:
      `(results, remaining_filters, remaining_orders)`.
      `results` is a list of entity_pb2.EntityProto
      remaining_filters and remaining_orders are the filters and orders that
      should be applied in memory
    """
    namespace_range = datastore_stub_util.ParseNamespaceQuery(query, filters,
                                                              orders)
    app_str = query.app

    namespaces = set()

    for app_namespace, _ in self._stub._GetAllEntities():
      (app_id, namespace) = datastore_types.DecodeAppIdNamespace(app_namespace)
      if app_id == app_str and namespace_range.Contains(namespace):
        namespaces.add(namespace)

    namespace_entities = []
    for namespace in namespaces:
      if namespace:
        namespace_e = datastore.Entity(
            self.name, name=namespace, _app=query.app)
      else:
        namespace_e = datastore.Entity(
            self.name, id=datastore_types._EMPTY_NAMESPACE_ID, _app=query.app)
      namespace_entities.append(namespace_e._ToPb())
    return (namespace_entities, [], [])


class DatastoreFileStub(datastore_stub_util.BaseDatastore,
                        apiproxy_stub.APIProxyStub,
                        datastore_stub_util.DatastoreStub):
  """Persistent stub for the Python datastore API.

  Stores all entities in memory, and persists them to a file as pickled
  protocol buffers. A `DatastoreFileStub` instance handles a single app's data
  and is backed by files on disk.
  """

  def __init__(self,
               app_id,
               datastore_file,
               history_file=None,
               require_indexes=False,
               service_name='datastore_v3',
               trusted=False,
               consistency_policy=None,
               save_changes=True,
               root_path=None,
               use_atexit=True,
               auto_id_policy=datastore_stub_util.SEQUENTIAL):
    """Constructor.

    Initializes and loads the datastore from the backing files, if they exist.

    Args:
      app_id: String
      datastore_file: String. Stores all entities across sessions. Use `None`
          not to use a file.
      history_file: DEPRECATE`. No-op.
      require_indexes: Bool. Default is `False`. If `True`, composite indexes
          must exist in `index.yaml` for queries that need them.
      service_name: Service name is expected for all calls.
      trusted: Bool. Default is `False`. If `True`, this stub allows an app to
        access the data of another app.
      consistency_policy: The consistency policy to use or `None` to use the
        default. Consistency policies can be found in
        `datastore_stub_util.*ConsistencyPolicy`.
      save_changes: Bool. Default is `True`. If this stub should modify
        `datastore_file` when entities are changed.
      root_path: String. The root path of the app.
      use_atexit: Bool. Indicates if the stub should save itself atexit.
      auto_id_policy: Enum. `datastore_stub_util.SEQUENTIAL` or `.SCATTERED`.
    """





    self.__datastore_file = datastore_file
    self.__save_changes = save_changes








    self.__entities_by_kind = collections.defaultdict(dict)
    self.__entities_by_group = collections.defaultdict(dict)
    self.__entities_lock = threading.Lock()




    self.__schema_cache = {}

    self.__id_counters = {
        datastore_stub_util.SEQUENTIAL: long(1),
        datastore_stub_util.SCATTERED: long(1)
    }
    self.__id_lock = threading.Lock()

    self.__file_lock = threading.Lock()

    datastore_stub_util.BaseDatastore.__init__(
        self, require_indexes, consistency_policy,
        use_atexit and self.__IsSaveable(), auto_id_policy)
    apiproxy_stub.APIProxyStub.__init__(self, service_name)
    datastore_stub_util.DatastoreStub.__init__(self, weakref.proxy(self),
                                               app_id, trusted, root_path)


    self._RegisterPseudoKind(KindPseudoKind())
    self._RegisterPseudoKind(PropertyPseudoKind())
    self._RegisterPseudoKind(NamespacePseudoKind())
    self._RegisterPseudoKind(datastore_stub_util.EntityGroupPseudoKind())

    self.Read()

  def Clear(self):
    """Clears the datastore by deleting all currently stored entities and queries.
    """
    self.__entities_lock.acquire()
    try:
      datastore_stub_util.BaseDatastore.Clear(self)
      datastore_stub_util.DatastoreStub.Clear(self)

      self.__entities_by_kind = collections.defaultdict(dict)
      self.__entities_by_group = collections.defaultdict(dict)
      self.__schema_cache = {}
    finally:
      self.__entities_lock.release()

  def _GetAllEntities(self):
    """Get all entities.

    Returns:
      Map from kind to `_StoredEntity()` list. Do not modify directly.
    """
    return self.__entities_by_kind

  def _GetEntityLocation(self, key):
    """Get keys to self.__entities_by_* from the given key.

    Example usage:
      app_kind, eg_k, k = self._GetEntityLocation(key)
      self.__entities_by_kind[app_kind][k]
      self.__entities_by_entity_group[eg_k][k]

    Args:
      key: entity_pb2.Reference

    Returns:
      Tuple (by_kind key, by_entity_group key, entity key)
    """
    app_ns = datastore_types.EncodeAppIdNamespace(key.app, key.name_space)
    kind = _FinalElement(key).type
    entity_group = datastore_stub_util._GetEntityGroup(key)
    eg_k = datastore_types.ReferenceToKeyValue(entity_group)
    k = datastore_types.ReferenceToKeyValue(key)

    return ((app_ns, kind), eg_k, k)

  def _StoreRecord(self, record, insert=False):
    """ Store the given entity record.

    Any needed locking should be managed by the caller.

    Args:
      record: The `EntityRecord` to store.
      insert: If we should check for existence.
    """
    app_kind, eg_k, k = self._GetEntityLocation(record.entity.key)

    assert not insert or k not in self.__entities_by_kind[app_kind]

    stored_entity = _StoredEntity(record)
    self.__entities_by_kind[app_kind][k] = stored_entity
    self.__entities_by_group[eg_k][k] = stored_entity


    if app_kind in self.__schema_cache:
      del self.__schema_cache[app_kind]

  READ_PB_EXCEPTIONS = (message.DecodeError, LookupError, TypeError, ValueError)
  READ_ERROR_MSG = ('Data in %s is corrupt or a different version. '
                    'Try running with the --clear_datastore flag.\n%r')
  READ_PY250_MSG = ('Are you using FloatProperty and/or GeoPtProperty? '
                    'Unfortunately loading float values from the datastore '
                    'file does not work with Python 2.5.0. '
                    'Please upgrade to a newer Python 2.5 release or use '
                    'the --clear_datastore flag.\n')

  def Read(self):
    """Reads the datastore and history files into memory.

    The in-memory query history is cleared, but the datastore is *not*
    cleared; the entities in the files are merged into the entities in memory.
    If you want them to overwrite the in-memory datastore, call `Clear()` before
    calling `Read()`.

    If the datastore file contains an entity with the same app name, kind, and
    key as an entity already in the datastore, the entity from the file
    overwrites the entity in the datastore.

    Also sets each `ID` counter to one greater than the highest `ID` allocated
    so far in that counter's `ID` space.
    """
    if self.__datastore_file and self.__datastore_file != '/dev/null':
      entities = self.__ReadPickled(self.__datastore_file)
      if entities and isinstance(entities[-1], six.integer_types):
        self._commit_timestamp = int(entities.pop())
      for encoded_entity in entities:
        try:
          entity = entity_pb2.EntityProto.FromString(encoded_entity)
          record = datastore_stub_util._FromStorageEntity(entity)
        except self.READ_PB_EXCEPTIONS as e:
          raise apiproxy_errors.ApplicationError(
              datastore_pb.Error.INTERNAL_ERROR,
              self.READ_ERROR_MSG % (self.__datastore_file, e))
        except struct.error as e:
          raise

        self._StoreRecord(record)

        last_path = _FinalElement(entity.key)
        if last_path.id:
          self._SetMaxId(last_path.id)

  def Write(self):
    """Writes out the datastore and history files.

    Important: If the files already exist, this method overwrites them.
    """
    super(DatastoreFileStub, self).Write()
    self.__WriteDatastore()

  def __IsSaveable(self):
    return (self.__datastore_file and self.__datastore_file != '/dev/null' and
            self.__save_changes)

  def __WriteDatastore(self):
    """Writes out the datastore file.

    Be careful: If the file already exists, this method overwrites it!
    """
    if self.__IsSaveable():
      encoded = []
      for kind_dict in self.__entities_by_kind.values():
        encoded.extend(entity.encoded_protobuf for entity in kind_dict.values())
      encoded.append(self._GetReadTimestamp())

      self.__WritePickled(encoded, self.__datastore_file)

  def __ReadPickled(self, filename):
    """Reads a pickled object from the given file and returns it.
    """
    self.__file_lock.acquire()

    try:
      try:
        if (filename and
            filename != '/dev/null' and
            os.path.isfile(filename) and
            os.stat(filename).st_size > 0):
          pickle_kwargs = {}
          if six.PY3:

            pickle_kwargs['encoding'] = 'bytes'
          return pickle.load(open(filename, 'rb'), **pickle_kwargs)
        else:
          logger.warning('Could not read datastore data from %s', filename)
      except (AttributeError, LookupError, ImportError, NameError, TypeError,
              ValueError, struct.error, pickle.PickleError) as e:


        raise apiproxy_errors.ApplicationError(
            datastore_pb.Error.INTERNAL_ERROR,
            'Could not read data from %s. Try running with the '
            '--clear_datastore flag. Cause:\n%r' % (filename, e))
    finally:
      self.__file_lock.release()

    return []

  def __WritePickled(self, obj, filename):
    """Pickles the object and writes it to the given file.
    """
    if not filename or filename == '/dev/null' or not obj:
      return


    descriptor, tmp_filename = tempfile.mkstemp(dir=os.path.dirname(filename))
    tmpfile = os.fdopen(descriptor, 'wb')





    pickler = pickle.Pickler(tmpfile, protocol=1)
    pickler.fast = True
    pickler.dump(obj)

    tmpfile.close()

    self.__file_lock.acquire()
    try:
      try:

        os.rename(tmp_filename, filename)
      except OSError:

        try:
          os.remove(filename)
        except:
          pass
        os.rename(tmp_filename, filename)
    finally:
      self.__file_lock.release()

  def MakeSyncCall(self, service, call, request, response, request_id=None):
    """The main RPC entry point. Service must be datastore_v3."""
    self.assertPbIsInitialized(request)
    super(DatastoreFileStub, self).MakeSyncCall(service,
                                                call,
                                                request,
                                                response,
                                                request_id)
    self.assertPbIsInitialized(response)

  def assertPbIsInitialized(self, pb):
    """Raises an exception if the given `PB` is not initialized and valid."""
    explanation = []

    pb.SerializeToString()

  def _GetSchemaCache(self, kind, usekey):
    if kind in self.__schema_cache and usekey in self.__schema_cache[kind]:
      return self.__schema_cache[kind][usekey]
    else:
      return None

  def _SetSchemaCache(self, kind, usekey, value):
    if kind not in self.__schema_cache:
      self.__schema_cache[kind] = {}
    self.__schema_cache[kind][usekey] = value



  def _Put(self, record, insert):
    record = datastore_stub_util.StoreRecord(record)

    self.__entities_lock.acquire()
    try:
      self._StoreRecord(record, insert)
    finally:
      self.__entities_lock.release()

  def _Get(self, key):
    app_kind, _, k = self._GetEntityLocation(key)


    try:
      stored_entity = self.__entities_by_kind[app_kind][k]
      return datastore_stub_util.LoadRecord(stored_entity.record)

    except KeyError:
      pass

  def _Delete(self, key):
    app_kind, eg_k, k = self._GetEntityLocation(key)

    self.__entities_lock.acquire()
    try:
      del self.__entities_by_kind[app_kind][k]
      del self.__entities_by_group[eg_k][k]
      if not self.__entities_by_kind[app_kind]:

        del self.__entities_by_kind[app_kind]
      if not self.__entities_by_group[eg_k]:
        del self.__entities_by_group[eg_k]

      del self.__schema_cache[app_kind]
    except KeyError:

      pass
    finally:
      self.__entities_lock.release()

  def _GetEntitiesInEntityGroup(self, entity_group):
    eg_k = datastore_types.ReferenceToKeyValue(entity_group)
    return dict((k, e.record)
                for (k, e) in six.iteritems(self.__entities_by_group[eg_k]))

  def _GetQueryCursor(self, query, filters, orders, index_list):
    app_id = query.app
    namespace = query.name_space

    pseudo_kind = None
    if query.HasField('kind') and query.kind in self._pseudo_kinds:
      pseudo_kind = self._pseudo_kinds[query.kind]




    self.__entities_lock.acquire()
    try:
      app_ns = datastore_types.EncodeAppIdNamespace(app_id, namespace)
      if pseudo_kind:

        (results, filters, orders) = pseudo_kind.Query(query, filters, orders)
        results = list(map(datastore_stub_util.EntityRecord, results))
      elif query.HasField('kind'):
        stored_entities = list(
            self.__entities_by_kind[app_ns, query.kind].values())
        results = [stored_entity.record for stored_entity in stored_entities]
      else:
        results = []
        for (cur_app_ns, _), entities in six.iteritems(self.__entities_by_kind):
          if cur_app_ns == app_ns:
            results.extend(entity.record for entity in six.itervalues(entities))
    except KeyError:
      results = []
    finally:
      self.__entities_lock.release()

    return datastore_stub_util._ExecuteQuery(results, query, filters, orders,
                                             index_list)

  def _SetIdCounter(self, id_space, value):
    """Sets the `ID` counter for id_space to value."""
    self.__id_counters[id_space] = value

  def _IdCounter(self, id_space):
    """Returns current value of `ID` counter for id_space."""
    return self.__id_counters[id_space]

  def _SetMaxId(self, max_id):
    """Infers `ID` space and advance corresponding counter."""
    count, id_space = datastore_stub_util.IdToCounter(max_id)
    if count >= self._IdCounter(id_space):
      self._SetIdCounter(id_space, count + 1)

  def _AllocateSequentialIds(self, reference, size=1, max_id=None):
    datastore_stub_util.Check(not (size and max_id),
                              'Both size and max cannot be set.')

    self.__id_lock.acquire()
    try:
      id_space = datastore_stub_util.SEQUENTIAL
      start = self._IdCounter(id_space)
      if size:
        datastore_stub_util.Check(size > 0, 'Size must be greater than 0.')
        self._SetIdCounter(id_space, start + size)
      elif max_id:
        datastore_stub_util.Check(max_id >=0,
                                  'Max must be greater than or equal to 0.')
        self._SetIdCounter(id_space, max(start, max_id + 1))
      end = self._IdCounter(id_space) - 1
    finally:
      self.__id_lock.release()

    return (start, end)

  def _AllocateIds(self, keys):
    self.__id_lock.acquire()
    full_keys = []
    try:
      for key in keys:
        last_element = _FinalElement(key)

        if last_element.id or last_element.HasField('name'):
          for el in key.path.element:
            if el.HasField('id'):
              self._SetMaxId(el.id)

        else:
          id_space = datastore_stub_util.SCATTERED
          count = self._IdCounter(id_space)
          last_element.id = datastore_stub_util.ToScatteredId(count)
          self._SetIdCounter(id_space, count + 1)
          full_keys.append(key)
    finally:
      self.__id_lock.release()

    return full_keys



#!/usr/bin/env python
#
# Copyright 2007 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


"""Stub version of the capability service API."""






import copy

from google.appengine.api import apiproxy_stub
from google.appengine.api import capabilities as capabilities_lib



IsEnabledRequest = capabilities_lib.IsEnabledRequest
IsEnabledResponse = capabilities_lib.IsEnabledResponse
CapabilityConfig = capabilities_lib.CapabilityConfig




_SUPPORTED_CAPABILITIES = {
    'blobstore': {'*': CapabilityConfig.ENABLED},
    'datastore_v3': {'*': CapabilityConfig.ENABLED,
                     'write': CapabilityConfig.ENABLED},
    'images': {'*': CapabilityConfig.ENABLED},
    'mail': {'*': CapabilityConfig.ENABLED},
    'memcache': {'*': CapabilityConfig.ENABLED},
    'taskqueue': {'*': CapabilityConfig.ENABLED},
    'urlfetch': {'*': CapabilityConfig.ENABLED},
    'xmpp': {'*': CapabilityConfig.ENABLED}
}




_CONFIG_STATUS_TO_SUMMARY_STATUS = {
    CapabilityConfig.DISABLED: IsEnabledResponse.DISABLED,
    CapabilityConfig.ENABLED: IsEnabledResponse.ENABLED,
    CapabilityConfig.SCHEDULED: IsEnabledResponse.SCHEDULED_NOW,
    CapabilityConfig.UNKNOWN: IsEnabledResponse.UNKNOWN
}


class CapabilityServiceStub(apiproxy_stub.APIProxyStub):
  """Capability service stub."""

  THREADSAFE = True

  def __init__(self, service_name='capability_service'):
    """Constructor.

    Args:
      service_name: Service name expected for all calls.
    """
    super(CapabilityServiceStub, self).__init__(service_name)
    self._packages = copy.deepcopy(_SUPPORTED_CAPABILITIES)

  def SetPackageEnabled(self, package, enabled):
    """Set all features of a given package to enabled.

    This method is thread-unsafe, so should only be called during set-up, before
    multiple API server threads start.

    Args:
      package: Name of package.
      enabled: True to enable, False to disable.

    Raises:
      KeyError: When an unsupported package is requested.
    """
    if package not in _SUPPORTED_CAPABILITIES:


      raise KeyError('Unsupported package. Received "%s". Must be one of %s.' %
                     (package, list(self._packages.values())))

    status = CapabilityConfig.ENABLED if enabled else CapabilityConfig.DISABLED
    for capability in self._packages.get(package):
      self.SetCapabilityStatus(package, capability, status)

  def SetCapabilityStatus(self, package, capability, status):
    """Set the status of an individual capability.

    This method is thread-unsafe, so should only be called during set-up, before
    multiple API server threads start.

    Args:
      package: String name of package containing the capability.
      capability: String capability name.
      status: The CapabilityConfig status enum to set.

    Raises:
      KeyError: When an unsupported package or capability is requested.
    """


    if package not in _SUPPORTED_CAPABILITIES:
      raise KeyError('Unsupported package. Received "%s". Must be one of %s.' %
                     (package, list(self._packages.values())))

    if capability not in _SUPPORTED_CAPABILITIES[package]:
      raise KeyError(
          'Unsupported capability for package "%s". Received "%s". Must be one '
          'of %s.' %
          (package, capability, list(self._packages[package].values())))

    self._packages[package][capability] = status

  def _Dynamic_IsEnabled(self, request, response):
    """Implementation of CapabilityService::IsEnabled().

    Args:
      request: An IsEnabledRequest.
      response: An IsEnabledResponse.
    """
    capability_statuses = []
    for capability in request.capability:
      default_config = response.config.add()
      default_config.package = request.package
      default_config.capability = capability
      try:
        config_status = self._packages[request.package][capability]
      except KeyError:
        config_status = CapabilityConfig.UNKNOWN
      default_config.status = config_status
      capability_statuses.append(config_status)




    summary_status = _CONFIG_STATUS_TO_SUMMARY_STATUS[max(capability_statuses)]
    response.summary_status = summary_status

  def _Dynamic_SetCapabilityStatus(self, request, response):
    """Implementation of CapabilityStubService::SetCapabilityStatus().

    Args:
      request: A capability_stub_service_pb2.SetCapabilityStatusRequest.
      response: A capability_stub_service_pb2.SetCapabilityStatusResponse.
    """
    self.SetCapabilityStatus(
        request.service_name, request.call, request.status)

  def Clear(self):
    self._packages = copy.deepcopy(_SUPPORTED_CAPABILITIES)



#!/usr/bin/env python
#
# Copyright 2007 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#














"""Polymorphic models and queries.

The standard NDB Model class only supports 'functional polymorphism'.
That is, you can create a subclass of Model, and then subclass that
class, as many generations as necessary, and those classes will share
all the same properties and behaviors of their base classes.  However,
subclassing Model in this way gives each subclass its own kind.  This
means that it is not possible to do 'polymorphic queries'.  Building a
query on a base class will only return entities whose kind matches
that base class's kind, and exclude entities that are instances of
some subclass of that base class.

The PolyModel class defined here lets you create class hierarchies
that support polymorphic queries.  Simply subclass PolyModel instead
of Model.
"""

import six
from google.appengine.ext.ndb import model

__all__ = ['PolyModel']

_CLASS_KEY_PROPERTY = 'class'


class _ClassKeyProperty(model.StringProperty):
  """Property to store the 'class key' of a polymorphic class.

  The class key is a list of strings describing a polymorphic entity's
  place within its class hierarchy.  This property is automatically
  calculated.  For example::

    class Foo(PolyModel): ...
    class Bar(Foo): ...
    class Baz(Bar): ...

    Foo().class_ == ['Foo']
    Bar().class_ == ['Foo', 'Bar']
    Baz().class_ == ['Foo', 'Bar', 'Baz']

  """

  def __init__(self, name=_CLASS_KEY_PROPERTY, indexed=True):
    """Constructor.

    If you really want to you can give this a different datastore name
    or make it unindexed.  For example::

      class Foo(PolyModel):
        class_ = _ClassKeyProperty(indexed=False)

    """
    super(_ClassKeyProperty, self).__init__(name=name, indexed=indexed,
                                            repeated=True)

  def _set_value(self, entity, value):
    """The class_ property is read-only from the user's perspective."""
    raise TypeError('%s is a read-only property' % self._code_name)

  def _get_value(self, entity):
    """Compute and store a default value if necessary."""
    value = super(_ClassKeyProperty, self)._get_value(entity)
    if not value:
      value = entity._class_key()
      self._store_value(entity, value)
    return value

  def _prepare_for_put(self, entity):
    """Ensure the class_ property is initialized before it is serialized."""
    self._get_value(entity)


class PolyModel(model.Model):
  """Base class for class hierarchies supporting polymorphic queries.

  Use this class to build hierarchies that can be queried based on
  their types.

  Example:

    Consider the following model hierarchy::

      +------+
      |Animal|
      +------+
        |
        +-----------------+
        |                 |
      +------+          +------+
      |Canine|          |Feline|
      +------+          +------+
        |                 |
        +-------+         +-------+
        |       |         |       |
      +---+   +----+    +---+   +-------+
      |Dog|   |Wolf|    |Cat|   |Panther|
      +---+   +----+    +---+   +-------+

    This class hierarchy has three levels.  The first is the 'root
    class'.  All models in a single class hierarchy must inherit from
    this root.  All models in the hierarchy are stored as the same
    kind as the root class.  For example, Panther entities when stored
    to Cloud Datastore are of the kind 'Animal'.  Querying against the
    Animal kind will retrieve Cats, Dogs and Canines, for example,
    that match your query.  Different classes stored in the root
    class' kind are identified by their class key.  When loaded from
    Cloud Datastore, it is mapped to the appropriate implementation
    class.

  Polymorphic properties:

    Properties that are defined in a given base class within a
    hierarchy are stored in Cloud Datastore for all subclasses only.
    So, if the Feline class had a property called 'whiskers', the Cat
    and Panther entities would also have whiskers, but not Animal,
    Canine, Dog or Wolf.

  Polymorphic queries:

    When written to Cloud Datastore, all polymorphic objects
    automatically have a property called 'class' that you can query
    against.  Using this property it is possible to easily write a
    query against any sub-hierarchy.  For example, to fetch only
    Canine objects, including all Dogs and Wolves::

      Canine.query()

    The 'class' property is not meant to be used by your code other
    than for queries.  Since it is supposed to represents the real
    Python class it is intended to be hidden from view.  (Although if
    you feel the need, it is accessible as the 'class_' attribute.)

  Root class:

    The root class is the class from which all other classes of the
    hierarchy inherits from.  Each hierarchy has a single root class.
    A class is a root class if it is an immediate child of PolyModel.
    The subclasses of the root class are all the same kind as the root
    class. In other words::

      Animal.kind() == Feline.kind() == Panther.kind() == 'Animal'

  Note:

    All classes in a given hierarchy must have unique names, since
    the class name is used to identify the appropriate subclass.
  """

  class_ = _ClassKeyProperty()

  _class_map = {}

  @classmethod
  def _update_kind_map(cls):
    """Override; called by Model._fix_up_properties().

    Update the kind map as well as the class map, except for PolyModel
    itself (its class key is empty).  Note that the kind map will
    contain entries for all classes in a PolyModel hierarchy; they all
    have the same kind, but different class names.  PolyModel class
    names, like regular Model class names, must be globally unique.
    """
    cls._kind_map[cls._class_name()] = cls
    class_key = cls._class_key()
    if class_key:
      cls._class_map[tuple(class_key)] = cls

  @classmethod
  def _from_pb(cls, pb, set_key=True, ent=None, key=None):
    """Override.

    Use the class map to give the entity the correct subclass.
    """
    prop_name = cls.class_._name
    class_name = []
    for plist in [pb.property, pb.raw_property]:
      for p in plist:
        if six.ensure_binary(p.name) == prop_name:
          class_name.append(p.value.stringValue)
    cls = cls._class_map.get(tuple(class_name), cls)
    return super(PolyModel, cls)._from_pb(pb, set_key, ent, key)

  @classmethod
  def _class_key(cls):
    """Return the class key.

    This is a list of class names, e.g. ['Animal', 'Feline', 'Cat'].
    """
    return [six.ensure_binary(c._class_name()) for c in cls._get_hierarchy()]

  @classmethod
  def _get_kind(cls):
    """Override.

    Make sure that the kind returned is the root class of the
    polymorphic hierarchy.
    """
    bases = cls._get_hierarchy()
    if not bases:






      return model.Model._get_kind.__func__(cls)
    else:
      return bases[0]._class_name()

  @classmethod
  def _class_name(cls):
    """Return the class name.

    This overrides Model._class_name() which is an alias for _get_kind().

    This is overridable in case you want to use a different class
    name.  The main use case is probably to maintain backwards
    compatibility with datastore contents after renaming a class.

    NOTE: When overriding this for an intermediate class in your
    hierarchy (as opposed to a leaf class), make sure to test
    cls.__name__, or else all subclasses will appear to have the
    same class name.
    """
    return cls.__name__

  @classmethod
  def _get_hierarchy(cls):
    """Internal helper to return the list of polymorphic base classes.

    This returns a list of class objects, e.g. [Animal, Feline, Cat].
    """
    bases = []
    for base in cls.mro():
      if hasattr(base, '_get_hierarchy'):
        bases.append(base)
    del bases[-1]
    bases.reverse()
    return bases

  @classmethod
  def _default_filters(cls):
    if len(cls._get_hierarchy()) <= 1:
      return ()
    return (cls.class_ == cls._class_name(),)



#!/usr/bin/env python
#
# Copyright 2007 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


"""Datastore backed Blobstore API stub.

Class:
  BlobstoreServiceStub: BlobstoreService stub backed by datastore.
"""











import base64
import os
import tempfile
import time

import six
import six.moves.urllib.parse

from google.appengine.api import apiproxy_stub
from google.appengine.api import blobstore
from google.appengine.api import datastore
from google.appengine.api import datastore_errors
from google.appengine.api import datastore_types
from google.appengine.api import users
from google.appengine.api.blobstore import blobstore_service_pb2
from google.appengine.api.blobstore import blobstore_stub_service_pb2
from google.appengine.api.blobstore import dict_blob_storage
from google.appengine.api.blobstore import file_blob_storage
from google.appengine.runtime import apiproxy_errors


__all__ = ['BlobstoreServiceStub',
           'ConfigurationError',
           'CreateUploadSession',
           'Error',
          ]


class Error(Exception):
  """Base blobstore error type."""


class ConfigurationError(Error):
  """Raised when environment is not correctly configured."""


_UPLOAD_SESSION_KIND = '__BlobUploadSession__'

_GS_INFO_KIND = '__GsFileInfo__'


def CreateUploadSession(creation,
                        success_path,
                        user,
                        max_bytes_per_blob,
                        max_bytes_total,
                        bucket_name=None):
  """Create upload session in datastore.

  Creates an upload session and puts it in Datastore to be referenced by
  upload handler later.

  Args:
    creation: Creation timestamp.
    success_path: Path in users application to call upon success.
    user: User that initiated this upload, if any.
    max_bytes_per_blob: Maximum number of bytes for any blob in the upload.
    max_bytes_total: Maximum aggregate bytes for all blobs in the upload.
    bucket_name: Name of the Google Storage bucket tio upload the files.

  Returns:
    String encoded key of new Datastore entity.
  """
  entity = datastore.Entity(_UPLOAD_SESSION_KIND, namespace='')
  entity_dict = {'creation': creation,
                 'success_path': success_path,
                 'user': user,
                 'state': 'init',
                 'max_bytes_per_blob': max_bytes_per_blob,
                 'max_bytes_total': max_bytes_total}
  if bucket_name:
    entity_dict['gs_bucket_name'] = bucket_name

  entity.update(entity_dict)
  datastore.Put(entity)
  return str(entity.key())


class BlobstoreServiceStub(apiproxy_stub.APIProxyStub):
  """Datastore backed Blobstore service stub.

  This stub stores manages upload sessions in the Datastore and must be
  provided with a blob_storage object to know where the actual blob
  records can be found after having been uploaded.

  This stub does not handle the actual creation of blobs, neither the BlobInfo
  in the Datastore nor creation of blob data in the blob_storage.  It does,
  however, assume that another part of the system has created these and
  uses these objects for deletion.

  An upload session is created when the CreateUploadURL request is handled and
  put in the Datastore under the __BlobUploadSession__ kind.  There is no
  analog for this kind on a production server. Other than creation, this stub
  not work with session objects.  The URLs created by this service stub are:

    http://<appserver-host>:<appserver-port>/<uploader-path>/<session-info>

  This is very similar to what the URL is on a production server.  The session
  info is the string encoded version of the session entity
  """

  _ACCEPTS_REQUEST_ID = True
  GS_BLOBKEY_PREFIX = 'encoded_gs_file:'


  THREADSAFE = False

  def __init__(self,
               blob_storage,
               time_function=time.time,
               service_name='blobstore',
               uploader_path='_ah/upload/',
               request_data=None,
               storage_dir=None,
               app_id='app-id'):
    """Constructor.

    Args:
      blob_storage: BlobStorage class instance used for blob storage.
      time_function: Used for dependency injection in tests.
      service_name: Service name expected for all calls.
      uploader_path: Path to upload handler pointed to by URLs generated
        by this service stub.
      request_data: A apiproxy_stub.RequestData instance used to look up state
        associated with the request that generated an API call.
      storage_dir: Directory for blobstore file storage if clients sets
        no_storage to true.
      app_id: App ID for blobstore file storage.
    """
    super(BlobstoreServiceStub, self).__init__(service_name,
                                               request_data=request_data)
    self.__storage = blob_storage
    self.__time_function = time_function
    self.__next_session_id = 1
    self.__uploader_path = uploader_path

    if not storage_dir:
      storage_dir = os.path.join(tempfile.mkdtemp(), 'blobstore')
    self._storage_dir = storage_dir
    self._app_id = app_id

  @classmethod
  def ToDatastoreBlobKey(cls, blobkey):
    """Given a string blobkey, return its db.Key."""
    kind = blobstore.BLOB_INFO_KIND
    if blobkey.startswith(cls.GS_BLOBKEY_PREFIX):
      kind = _GS_INFO_KIND
    return datastore_types.Key.from_path(kind,
                                         blobkey,
                                         namespace='')

  @property
  def storage(self):
    """Access BlobStorage used by service stub.

    Returns:
      BlobStorage instance used by blobstore service stub.
    """
    return self.__storage

  def _CreateSession(self,
                     success_path,
                     user,
                     max_bytes_per_blob=None,
                     max_bytes_total=None,
                     bucket_name=None):
    """Create new upload session.

    Args:
      success_path: Application path to call upon successful POST.
      user: User that initiated the upload session.
      max_bytes_per_blob: Maximum number of bytes for any blob in the upload.
      max_bytes_total: Maximum aggregate bytes for all blobs in the upload.
      bucket_name: The name of the Cloud Storage bucket where the files will be
        uploaded.

    Returns:
      String encoded key of a new upload session created in the datastore.
    """
    return CreateUploadSession(self.__time_function(),
                               success_path,
                               user,
                               max_bytes_per_blob,
                               max_bytes_total,
                               bucket_name)

  def _Dynamic_CreateUploadURL(self, request, response, request_id):
    """Create upload URL implementation.

    Create a new upload session.  The upload session key is encoded in the
    resulting POST URL.  This URL is embedded in a POST form by the application
    which contacts the uploader when the user posts.

    Args:
      request: A fully initialized CreateUploadURLRequest instance.
      response: A CreateUploadURLResponse instance.
      request_id: A unique string identifying the request associated with the
          API call.
    """
    max_bytes_per_blob = None
    max_bytes_total = None
    bucket_name = None

    if request.HasField('max_upload_size_per_blob_bytes'):
      max_bytes_per_blob = request.max_upload_size_per_blob_bytes

    if request.HasField('max_upload_size_bytes'):
      max_bytes_total = request.max_upload_size_bytes

    if request.HasField('gs_bucket_name'):
      bucket_name = request.gs_bucket_name

    session = self._CreateSession(request.success_path,
                                  users.get_current_user(), max_bytes_per_blob,
                                  max_bytes_total, bucket_name)

    protocol, host, _, _, _, _ = six.moves.urllib.parse.urlparse(
        self.request_data.get_request_url(request_id))

    response.url = ('%s://%s/%s%s' %
                    (protocol, host, self.__uploader_path, session))

  @classmethod
  def DeleteBlob(cls, blobkey, storage):
    """Delete a blob.

    Args:
      blobkey: blobkey in str.
      storage: blobstore storage stub.
    """
    datastore.Delete(cls.ToDatastoreBlobKey(blobkey))

    blobinfo = datastore_types.Key.from_path(blobstore.BLOB_INFO_KIND,
                                             blobkey,
                                             namespace='')
    datastore.Delete(blobinfo)
    storage.DeleteBlob(blobkey)

  def _Dynamic_DeleteBlob(self, request, response, unused_request_id):
    """Delete a blob by its blob-key.

    Delete a blob from the blobstore using its blob-key.  Deleting blobs that
    do not exist is a no-op.

    Args:
      request: A fully initialized DeleteBlobRequest instance.
      response: Not used but should be a VoidProto.
    """
    for blobkey in request.blob_key:
      self.DeleteBlob(blobkey, self.__storage)

  def _Dynamic_StoreBlob(self, request, response, unused_request_id):
    """Directly store a blob in the local blobstore.

    This is defined on the BlobstoreStubService, and is intended for use in
    tests only. It was introduced primarily to support Java local development.

    Args:
      request: An instance of blobstore_stub_service_pb2.StoreBlobRequest.
      response: An unused instance of api_base_pb2.VoidProto.
    """
    del response
    del unused_request_id

    self.CreateBlob(request.blob_key, request.content)

  def _Dynamic_SetBlobStorageType(self, request, unused_response,
                                  unused_request_id):
    """Configure whether stub should store blobs in-memory or on-disk.

    This is primarily used by the Java blobstore client APIs to switch between
    memory and file storage.

    Args:
      request: An instance of
        blobstore_stub_service_pb2.SetBlobStorageTypeRequest.
    """
    if request.storage_type == (
        blobstore_stub_service_pb2.SetBlobStorageTypeRequest.MEMORY):
      self.__storage = dict_blob_storage.DictBlobStorage()
    elif request.storage_type == (
        blobstore_stub_service_pb2.SetBlobStorageTypeRequest.FILE):
      self.__storage = file_blob_storage.FileBlobStorage(
          self._storage_dir, self._app_id)

  def _Dynamic_FetchData(self, request, response, unused_request_id):
    """Fetch a blob fragment from a blob by its blob-key.

    Fetches a blob fragment using its blob-key.  Start index is inclusive,
    end index is inclusive.  Valid requests for information outside of
    the range of the blob return a partial string or empty string if entirely
    out of range.

    Args:
      request: A fully initialized FetchDataRequest instance.
      response: A FetchDataResponse instance.

    Raises:
      ApplicationError when application has the following errors:
        INDEX_OUT_OF_RANGE: Index is negative or end > start.
        BLOB_FETCH_SIZE_TOO_LARGE: Request blob fragment is larger than
          MAX_BLOB_FRAGMENT_SIZE.
        BLOB_NOT_FOUND: If invalid blob-key is provided or is not found.
    """

    start_index = request.start_index
    if start_index < 0:
      raise apiproxy_errors.ApplicationError(
          blobstore_service_pb2.BlobstoreServiceError.DATA_INDEX_OUT_OF_RANGE)


    end_index = request.end_index
    if end_index < start_index:
      raise apiproxy_errors.ApplicationError(
          blobstore_service_pb2.BlobstoreServiceError.DATA_INDEX_OUT_OF_RANGE)


    fetch_size = end_index - start_index + 1
    if fetch_size > blobstore.MAX_BLOB_FETCH_SIZE:
      raise apiproxy_errors.ApplicationError(
          blobstore_service_pb2.BlobstoreServiceError.BLOB_FETCH_SIZE_TOO_LARGE)


    blobkey = request.blob_key
    info_key = self.ToDatastoreBlobKey(blobkey)
    try:
      datastore.Get(info_key)
    except datastore_errors.EntityNotFoundError:
      raise apiproxy_errors.ApplicationError(
          blobstore_service_pb2.BlobstoreServiceError.BLOB_NOT_FOUND)


    blob_file = self.__storage.OpenBlob(blobkey)
    blob_file.seek(start_index)
    response.data = six.ensure_binary(blob_file.read(fetch_size))

  def _Dynamic_DecodeBlobKey(self, request, response, unused_request_id):
    """Decode a given blob key: data is simply base64-decoded.

    Args:
      request: A fully-initialized DecodeBlobKeyRequest instance
      response: A DecodeBlobKeyResponse instance.
    """
    for blob_key in request.blob_key:
      response.add_decoded(blob_key.decode('base64'))

  @classmethod
  def CreateEncodedGoogleStorageKey(cls, filename):
    """Create an encoded blob key that represents a Google Storage file.

    For now we'll just base64 encode the Google Storage filename, APIs that
    accept encoded blob keys will need to be able to support Google Storage
    files or blobstore files based on decoding this key.

    Any stub that creates GS files should use this function to convert
    a gs filename to a blobkey. The created blobkey should be used both
    as its _GS_FILE_INFO entity's key name and as the storage key to
    store its content in blobstore. This ensures the GS files created
    can be operated by other APIs.

    Note this encoding is easily reversible and is not encryption.

    Args:
      filename: gs filename of form 'bucket/filename'

    Returns:
      blobkey string of encoded filename.
    """
    suffix = base64.urlsafe_b64encode(six.ensure_binary(filename))
    return cls.GS_BLOBKEY_PREFIX + suffix.decode('utf-8')

  def _Dynamic_CreateEncodedGoogleStorageKey(self, request, response,
                                             unused_request_id):
    """Create an encoded blob key that represents a Google Storage file.

    For now we'll just base64 encode the Google Storage filename, APIs that
    accept encoded blob keys will need to be able to support Google Storage
    files or blobstore files based on decoding this key.

    Args:
      request: A fully-initialized CreateEncodedGoogleStorageKeyRequest
        instance.
      response: A CreateEncodedGoogleStorageKeyResponse instance.
    """
    filename = request.filename[len(blobstore.GS_PREFIX):]
    response.blob_key = self.CreateEncodedGoogleStorageKey(filename)

  def CreateBlob(self, blob_key, content):
    """Create new blob and put in storage and Datastore.

    This is useful in testing where you have access to the stub.

    Args:
      blob_key: String blob-key of new blob.
      content: Content of new blob as a string.

    Returns:
      New Datastore entity without blob meta-data fields.
    """
    entity = datastore.Entity(blobstore.BLOB_INFO_KIND,
                              name=blob_key, namespace='')
    entity['size'] = len(content)
    datastore.Put(entity)
    self.storage.StoreBlob(blob_key, six.BytesIO(content))
    return entity



#!/usr/bin/env python
#
# Copyright 2007 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


"""Blobstore API.

This module contains methods that you can use to interface with the Blobstore
API. The module defines a `db.Key`-like class that represents a blob key.
"""









import datetime
import time

import six

from google.appengine.api import api_base_pb2
from google.appengine.api import apiproxy_stub_map
from google.appengine.api import datastore
from google.appengine.api import datastore_errors
from google.appengine.api import datastore_types
from google.appengine.api.blobstore import blobstore_service_pb2
from google.appengine.runtime import apiproxy_errors


__all__ = ['BLOB_INFO_KIND',
           'BLOB_KEY_HEADER',
           'BLOB_MIGRATION_KIND',
           'BLOB_RANGE_HEADER',
           'MAX_BLOB_FETCH_SIZE',
           'UPLOAD_INFO_CREATION_HEADER',
           'CLOUD_STORAGE_OBJECT_HEADER',
           'GS_PREFIX',
           'BlobFetchSizeTooLargeError',
           'BlobKey',
           'BlobNotFoundError',
           'DataIndexOutOfRangeError',
           'PermissionDeniedError',
           'Error',
           'InternalError',
           'create_rpc',
           'create_upload_url',
           'create_upload_url_async',
           'delete',
           'delete_async',
           'fetch_data',
           'fetch_data_async',
           'create_gs_key',
           'create_gs_key_async',
          ]


BlobKey = datastore_types.BlobKey



BLOB_INFO_KIND = '__BlobInfo__'

BLOB_KEY_HEADER = 'X-AppEngine-BlobKey'

BLOB_MIGRATION_KIND = '__BlobMigration__'

BLOB_RANGE_HEADER = 'X-AppEngine-BlobRange'

MAX_BLOB_FETCH_SIZE = (1 << 20) - (1 << 15)

GS_PREFIX = '/gs/'



UPLOAD_INFO_CREATION_HEADER = 'X-AppEngine-Upload-Creation'
CLOUD_STORAGE_OBJECT_HEADER = 'X-AppEngine-Cloud-Storage-Object'
_BASE_CREATION_HEADER_FORMAT = '%Y-%m-%d %H:%M:%S'

class Error(Exception):
  """Base blobstore error type."""


class InternalError(Error):
  """An internal error occured."""


class BlobNotFoundError(Error):
  """The blob does not exist."""


class DataIndexOutOfRangeError(Error):
  """The indexes could not be accessed.

  The specified indexes were out of range or in the wrong order.
  """


class BlobFetchSizeTooLargeError(Error):
  """The block could not be fetched because it was too large."""


class _CreationFormatError(Error):
  """The creation date could not be parsed because the format is invalid."""


class PermissionDeniedError(Error):
  """The operation did not complete; review the permissions required."""


def _ToBlobstoreError(error):
  """Translates an application error to a datastore error, if possible.

  Args:
    error: An `ApplicationError` to translate.

  Returns:
    The `BlobstoreError` for the passed-in `ApplicationError`, if one is found.
    Otherwise, the original `ApplicationError` is returned.
  """
  error_map = {
      blobstore_service_pb2.BlobstoreServiceError.INTERNAL_ERROR:
          InternalError,
      blobstore_service_pb2.BlobstoreServiceError.BLOB_NOT_FOUND:
          BlobNotFoundError,
      blobstore_service_pb2.BlobstoreServiceError.DATA_INDEX_OUT_OF_RANGE:
          DataIndexOutOfRangeError,
      blobstore_service_pb2.BlobstoreServiceError.BLOB_FETCH_SIZE_TOO_LARGE:
          BlobFetchSizeTooLargeError,
      blobstore_service_pb2.BlobstoreServiceError.PERMISSION_DENIED:
          PermissionDeniedError,
  }
  desired_exc = error_map.get(error.application_error)
  return desired_exc(error.error_detail) if desired_exc else error


def _format_creation(stamp):
  """Formats an upload creation timestamp, including microseconds.

  Use this method to format a timestamp that includes microseconds on Python
  versions prior to 2.6.

  For earlier versions of Python, you cannot simply convert datetime objects to
  strings, because the microseconds are stripped from the format when they are
  set to 0.

  The upload creation date format will always contain microseconds
  padded out to 6 places.

  Args:
    stamp: A `datetime.datetime` object that you want to format.

  Returns:
    A formatted datetime object as Python 2.6 format `%Y-%m-%d %H:%M:%S.%f`.
  """
  return '%s.%06d' % (stamp.strftime(_BASE_CREATION_HEADER_FORMAT),
                      stamp.microsecond)


def _parse_creation(creation_string, field_name):
  """Parses the upload creation string from header format.

  This method parses the creation date of objects using the following format:
  `YYYY-mm-dd HH:MM:SS.ffffff`, where:
      - `YYYY`: Year
      - `mm`: Month (01-12)
      - `dd`: Day (01-31)
      - `HH`: Hour (00-24)
      - `MM`: Minute (00-59)
      - `SS`: Second (00-59)
      - `ffffff`: Microsecond

  Args:
    creation_string: The string of the creation date, properly formatted.
    field_name: The field for which you are attempting to parse the
        `creation_string`.

  Returns:
    A datetime object that is parsed from `creation_string`.

  Raises:
    `_CreationFormatError`: If the creation string is formatted incorrectly.
  """


  split_creation_string = creation_string.split('.', 1)
  if len(split_creation_string) != 2:
    raise _CreationFormatError(
        'Could not parse creation %s in field %s.' % (creation_string,
                                                      field_name))
  timestamp_string, microsecond = split_creation_string

  try:
    timestamp = time.strptime(timestamp_string,
                              _BASE_CREATION_HEADER_FORMAT)
    microsecond = int(microsecond)
  except ValueError:
    raise _CreationFormatError('Could not parse creation %s in field %s.'
                               % (creation_string, field_name))

  return datetime.datetime(*timestamp[:6] + tuple([microsecond]))


def create_rpc(deadline=None, callback=None):
  """Creates an RPC object to use with the Blobstore API.

  Args:
    deadline: Optional deadline in seconds for the operation; the default value
        is a system-specific deadline, typically 5 seconds.
    callback: Optional callable to invoke on completion.

  Returns:
    An `apiproxy_stub_map.UserRPC` object that is specialized for this service.
  """
  return apiproxy_stub_map.UserRPC('blobstore', deadline, callback)


def _make_async_call(rpc, method, request, response,
                     get_result_hook, user_data):
  """Makes an asynchronous API call."""
  if rpc is None:
    rpc = create_rpc()
  rpc.make_call(method, request, response, get_result_hook, user_data)
  return rpc


def _get_result_hook(rpc):
  """If there was an exception, raise it now."""
  try:
    rpc.check_success()
  except apiproxy_errors.ApplicationError as err:
    raise _ToBlobstoreError(err)
  hook = rpc.user_data
  return hook(rpc)


def create_upload_url(success_path,
                      max_bytes_per_blob=None,
                      max_bytes_total=None,
                      rpc=None,
                      gs_bucket_name=None):
  """Creates the upload URL for a POST form.

  Args:
    success_path: Path within application to call when a `POST` call is
        successful and the upload is complete.
    max_bytes_per_blob: The maximum size in bytes that any one blob in the
        upload can be, or `None` for no maximum size.
    max_bytes_total: The maximum size in bytes that the aggregate sizes of all
        of the blobs in the upload can be, or `None` for no maximum size.
    rpc: Optional UserRPC object.
    gs_bucket_name: The Google Cloud Storage bucket name to which the blobs
        should be uploaded. The application's service account must have the
        correct permissions to write to this bucket. The bucket name can be of
        the format `bucket/path/`, in which case the included path will be
        prepended to the uploaded object name.

  Returns:
    The upload URL.

  Raises:
    TypeError: If `max_bytes_per_blob` or `max_bytes_total` are not integral
        types.
    ValueError: If `max_bytes_per_blob` or `max_bytes_total` are not
        positive values.
  """
  rpc = create_upload_url_async(success_path,
                                max_bytes_per_blob=max_bytes_per_blob,
                                max_bytes_total=max_bytes_total,
                                rpc=rpc,
                                gs_bucket_name=gs_bucket_name)
  return rpc.get_result()


def create_upload_url_async(success_path,
                            max_bytes_per_blob=None,
                            max_bytes_total=None,
                            rpc=None,
                            gs_bucket_name=None):
  """Asynchronously creates the upload URL for a POST form.

  Args:
    success_path: The path within the application to call when a `POST` call is
        successful and the upload is complete.
    max_bytes_per_blob: The maximum size in bytes that any one blob in the
        upload can be, or `None` for no maximum size.
    max_bytes_total: The maximum size in bytes that the aggregate sizes of all
        of the blobs in the upload can be, or `None` for no maximum size.
    rpc: Optional UserRPC object.
    gs_bucket_name: The Google Cloud Storage bucket name to which the blobs
        should be uploaded. The application's service account must have the
        correct permissions to write to this bucket. The bucket name can be of
        the format `bucket/path/`, in which case the included path will be
        prepended to the uploaded object name.

  Returns:
    A UserRPC whose result will be the upload URL.

  Raises:
    TypeError: If `max_bytes_per_blob` or `max_bytes_total` are not integral
        types.
    ValueError: If `max_bytes_per_blob` or `max_bytes_total` are not
        positive values.
  """
  request = blobstore_service_pb2.CreateUploadURLRequest()
  response = blobstore_service_pb2.CreateUploadURLResponse()
  request.success_path = success_path

  if max_bytes_per_blob is not None:
    if not isinstance(max_bytes_per_blob, six.integer_types):
      raise TypeError('max_bytes_per_blob must be integer.')
    if max_bytes_per_blob < 1:
      raise ValueError('max_bytes_per_blob must be positive.')
    request.max_upload_size_per_blob_bytes = max_bytes_per_blob

  if max_bytes_total is not None:
    if not isinstance(max_bytes_total, six.integer_types):
      raise TypeError('max_bytes_total must be integer.')
    if max_bytes_total < 1:
      raise ValueError('max_bytes_total must be positive.')
    request.max_upload_size_bytes = max_bytes_total

  if (request.HasField('max_upload_size_bytes') and
      request.HasField('max_upload_size_per_blob_bytes')):
    if request.max_upload_size_bytes < request.max_upload_size_per_blob_bytes:
      raise ValueError('max_bytes_total can not be less'
                       ' than max_upload_size_per_blob_bytes')

  if gs_bucket_name is not None:
    if not isinstance(gs_bucket_name, six.string_types):
      raise TypeError('gs_bucket_name must be a string.')
    request.gs_bucket_name = gs_bucket_name

  return _make_async_call(rpc, 'CreateUploadURL', request, response,
                          _get_result_hook, lambda rpc: rpc.response.url)




def delete(blob_keys, rpc=None, _token=None):
  """Deletes a blob from Blobstore.

  Args:
    blob_keys: A single `BlobKey` instance or a list of blob keys. A blob key
        can be either a string or an instance of `BlobKey`.
    rpc: Optional UserRPC object.

  Returns:
    None.
  """



  rpc = delete_async(blob_keys, rpc, _token)
  return rpc.get_result()




def delete_async(blob_keys, rpc=None, _token=None):
  """Asynchronously deletes a blob from Blobstore.

  Args:
    blob_keys: A single `BlobKey` instance or a list of blob keys. A blob key
        can be either a string or an instance of `BlobKey`.
    rpc: Optional UserRPC object.

  Returns:
    A UserRPC, whose result will be `None`.
  """



  if isinstance(blob_keys, (six.string_types, BlobKey)):
    blob_keys = [blob_keys]
  request = blobstore_service_pb2.DeleteBlobRequest()
  for blob_key in blob_keys:
    request.blob_key.append(str(blob_key))
  if _token:
    request.token = _token
  response = api_base_pb2.VoidProto()

  return _make_async_call(rpc, 'DeleteBlob', request, response,
                          _get_result_hook, lambda rpc: None)


def fetch_data(blob_key, start_index, end_index, rpc=None):
  """Fetches the data for a blob.

  Args:
    blob_key: A `BlobKey`, string, or Unicode representation of a `BlobKey` of
        the blob from which you want to fetch data.
    start_index: The start index value of the blob data to fetch. This argument
        cannot be set to a negative value.
    end_index: The end index value (exclusive) of the blob data to fetch. This
        argument must be greater than or equal to the `start_index` value.
    rpc: Optional UserRPC object.

  Returns:
    A string containing partial data of the blob.
  """
  rpc = fetch_data_async(blob_key, start_index, end_index, rpc)
  return rpc.get_result()


def fetch_data_async(blob_key, start_index, end_index, rpc=None):
  """Asynchronously fetches the data for a blob.

  Args:
    blob_key: A `BlobKey`, string, or Unicode representation of a `BlobKey` of
        the blob from which you want to fetch data.
    start_index: The start index value of the blob data to fetch. This argument
        cannot be set to a negative value.
    end_index: The end index value (exclusive) of the blob data to fetch. This
        argument must be greater than or equal to the `start_index` value.
    rpc: Optional UserRPC object.

  Returns:
    A UserRPC whose result will be a string as returned by `fetch_data()`.
  """
  if not isinstance(start_index, six.integer_types):
    raise TypeError('start_index must be an integer.')

  if not isinstance(end_index, six.integer_types):
    raise TypeError('end_index must be an integer.')

  if isinstance(blob_key, BlobKey):
    blob_key = str(blob_key)
  elif isinstance(blob_key, str):
    blob_key = blob_key
  elif not isinstance(blob_key, six.text_type):
    raise TypeError('blob_key must be str, Unicode or BlobKey: %s' % blob_key)


  if start_index < 0:
    raise DataIndexOutOfRangeError(
        'Cannot fetch blob at negative index.')


  if end_index < start_index:
    raise DataIndexOutOfRangeError(
        'Start index %d > end index %d' % (start_index, end_index))


  fetch_size = end_index - start_index + 1

  if fetch_size > MAX_BLOB_FETCH_SIZE:
    raise BlobFetchSizeTooLargeError(
        'Blob fetch size is too large: %d' % fetch_size)

  request = blobstore_service_pb2.FetchDataRequest()
  response = blobstore_service_pb2.FetchDataResponse()

  request.blob_key = blob_key
  request.start_index = start_index
  request.end_index = end_index

  return _make_async_call(rpc, 'FetchData', request, response,
                          _get_result_hook, lambda rpc: rpc.response.data)


def create_gs_key(filename, rpc=None):
  """Creates an encoded key for a Google Cloud Storage file.

  It is safe to persist this key for future use.

  Args:
    filename: The file name of the Google Cloud Storage object for which you
        want to create the key.
    rpc: Optional UserRPC object.

  Returns:
    An encrypted `BlobKey` string.
  """
  rpc = create_gs_key_async(filename, rpc)
  return rpc.get_result()


def create_gs_key_async(filename, rpc=None):
  """Asynchronously creates an encoded key for a Google Cloud Storage file.

  It is safe to persist this key for future use.

  Args:
    filename: The file name of the Google Cloud Storage object for which you
        want to create the key.
    rpc: Optional UserRPC object.

  Returns:
    A UserRPC whose result will be a string as returned by `create_gs_key()`.

  Raises:
    TypeError: If `filename` is not a string.
    ValueError: If `filename` is not in the format
        `/gs/bucket_name/object_name`.
  """

  if not isinstance(filename, six.string_types):
    raise TypeError('filename must be str: %s' % filename)
  if not filename.startswith(GS_PREFIX):
    raise ValueError('filename must start with "/gs/": %s' % filename)
  if not '/' in filename[4:]:
    raise ValueError('filename must use the format '
                     '"/gs/bucket_name/object_name": %s' % filename)

  request = blobstore_service_pb2.CreateEncodedGoogleStorageKeyRequest()
  response = blobstore_service_pb2.CreateEncodedGoogleStorageKeyResponse()

  request.filename = filename

  return _make_async_call(rpc, 'CreateEncodedGoogleStorageKey', request,
                          response,
                          _get_result_hook, lambda rpc: rpc.response.blob_key)



#!/usr/bin/env python
#
# Copyright 2007 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Exposes methods to control services (modules) and versions of an app.

Services were formerly known as modules and the API methods still
reflect that naming. For more information and code samples, see
Using the Modules guide:
https://cloud.google.com/appengine/docs/standard/python/using-the-modules-api.
"""

__all__ = [
    'Error',
    'InvalidModuleError',
    'InvalidVersionError',
    'InvalidInstancesError',
    'UnexpectedStateError',
    'TransientError',

    'get_current_module_name',
    'get_current_version_name',
    'get_current_instance_id',
    'get_modules',
    'get_versions',
    'get_default_version',
    'get_num_instances',
    'set_num_instances',
    'set_num_instances_async',
    'start_version',
    'start_version_async',
    'stop_version',
    'stop_version_async',
    'get_hostname'
]

import logging
import os

import six

from google.appengine.api import apiproxy_stub_map
from google.appengine.api.modules import modules_service_pb2
from google.appengine.runtime import apiproxy_errors


class Error(Exception):
  """Base-class for errors in this module."""


class InvalidModuleError(Error):
  """The given module is not known to the system."""


class InvalidVersionError(Error):
  """The given module version is not known to the system."""


class InvalidInstancesError(Error):
  """The given instances value is not valid."""


class UnexpectedStateError(Error):
  """An unexpected current state was found when starting/stopping a module."""


class TransientError(Error):
  """A transient error was encountered, retry the operation."""


def get_current_module_name():
  """Returns the module name of the current instance.

  If this is version "v1" of module "module5" for app "my-app", this function
  will return "module5".
  """
  return os.environ.get('GAE_SERVICE') or os.environ.get('CURRENT_MODULE_ID')


def get_current_version_name():
  """Returns the version of the current instance.

  If this is version "v1" of module "module5" for app "my-app", this function
  will return "v1".
  """
  result = os.environ.get('GAE_VERSION')
  if result:
    return result

  result = os.environ['CURRENT_VERSION_ID'].split('.')[0]
  return None if result == 'None' else result


def get_current_instance_id():
  """Returns the ID of the current instance.

  If this is instance 2 of version "v1" of module "module5" for app "my-app",
  this function will return "2".

  This is only valid for automatically-scaled modules; otherwise, `None` is
  returned.


  Returns:
    String containing the ID of the instance, or `None` if this is not an
    automatically-scaled module.
  """
  return os.environ.get('GAE_INSTANCE') or os.environ.get('INSTANCE_ID', None)


def _GetRpc():
  return apiproxy_stub_map.UserRPC('modules')


def _MakeAsyncCall(method, request, response, get_result_hook):
  rpc = _GetRpc()
  rpc.make_call(method, request, response, get_result_hook)
  return rpc


_MODULE_SERVICE_ERROR_MAP = {
    modules_service_pb2.ModulesServiceError.INVALID_INSTANCES:
        InvalidInstancesError,
    modules_service_pb2.ModulesServiceError.INVALID_MODULE:
        InvalidModuleError,
    modules_service_pb2.ModulesServiceError.INVALID_VERSION:
        InvalidVersionError,
    modules_service_pb2.ModulesServiceError.TRANSIENT_ERROR:
        TransientError,
    modules_service_pb2.ModulesServiceError.UNEXPECTED_STATE:
        UnexpectedStateError
}


def _CheckAsyncResult(rpc, expected_application_errors,
                      ignored_application_errors):
  try:
    rpc.check_success()
  except apiproxy_errors.ApplicationError as e:
    if e.application_error in ignored_application_errors:
      logging.info(ignored_application_errors.get(e.application_error))
      return
    if e.application_error in expected_application_errors:
      mapped_error = _MODULE_SERVICE_ERROR_MAP.get(e.application_error)
      if mapped_error:
        raise mapped_error()
    raise Error(e)


def get_modules():
  """Returns a list of all modules for the application.

  Returns:
    List of strings containing the names of modules associated with this
      application.  The 'default' module will be included if it exists, as will
      the name of the module that is associated with the instance that calls
      this function.
  """

  def _ResultHook(rpc):
    _CheckAsyncResult(rpc, [], {})


    return rpc.response.module

  request = modules_service_pb2.GetModulesRequest()
  response = modules_service_pb2.GetModulesResponse()
  return _MakeAsyncCall('GetModules', request, response,
                        _ResultHook).get_result()


def get_versions(module=None):
  """Returns a list of versions for a given module.

  Args:
    module: Module to retrieve version for, if `None` then the current module
      will be used.

  Returns:
    List of strings containing the names of versions associated with the module.
    The current version will also be included in this list.

  Raises:
    `InvalidModuleError` if the given module isn't valid, `TransientError` if
    there is an issue fetching the information.
  """

  def _ResultHook(rpc):
    mapped_errors = [
        modules_service_pb2.ModulesServiceError.INVALID_MODULE,
        modules_service_pb2.ModulesServiceError.TRANSIENT_ERROR
    ]
    _CheckAsyncResult(rpc, mapped_errors, {})


    return rpc.response.version

  request = modules_service_pb2.GetVersionsRequest()
  if module:
    request.module = module
  response = modules_service_pb2.GetVersionsResponse()
  return _MakeAsyncCall('GetVersions', request, response,
                        _ResultHook).get_result()


def get_default_version(module=None):
  """Returns the name of the default version for the module.

  Args:
    module: Module to retrieve the default version for, if `None` then the
      current module will be used.

  Returns:
    String containing the name of the default version of the module.

  Raises:
    `InvalidModuleError` if the given module is not valid, `InvalidVersionError`
    if no default version could be found.
  """

  def _ResultHook(rpc):
    mapped_errors = [
        modules_service_pb2.ModulesServiceError.INVALID_MODULE,
        modules_service_pb2.ModulesServiceError.INVALID_VERSION
    ]
    _CheckAsyncResult(rpc, mapped_errors, {})
    return rpc.response.version

  request = modules_service_pb2.GetDefaultVersionRequest()
  if module:
    request.module = module
  response = modules_service_pb2.GetDefaultVersionResponse()
  return _MakeAsyncCall('GetDefaultVersion', request, response,
                        _ResultHook).get_result()


def get_num_instances(
    module=None,
    version=None):
  """Return the number of instances that are set for the given module version.

  This is only valid for fixed modules, an error will be raised for
  automatically-scaled modules.  Support for automatically-scaled modules may be
  supported in the future.

  Args:
    module: String containing the name of the module to fetch this info for, if
      `None` the module of the current instance will be used.
    version: String containing the name of the version to fetch this info for,
      if `None` the version of the current instance will be used.  If that
      version does not exist in the other module, then an `InvalidVersionError`
      is raised.

  Returns:
    The number of instances that are set for the given module version.

  Raises:
    `InvalidVersionError` on invalid input.
  """

  def _ResultHook(rpc):
    mapped_errors = [modules_service_pb2.ModulesServiceError.INVALID_VERSION]
    _CheckAsyncResult(rpc, mapped_errors, {})
    return rpc.response.instances

  request = modules_service_pb2.GetNumInstancesRequest()
  if module:
    request.module = module
  if version:
    request.version = version
  response = modules_service_pb2.GetNumInstancesResponse()
  return _MakeAsyncCall('GetNumInstances', request, response,
                        _ResultHook).get_result()


def set_num_instances(
    instances,
    module=None,
    version=None):
  """Sets the number of instances on the module and version.

  Args:
    instances: The number of instances to set.
    module: The module to set the number of instances for, if `None` the current
      module will be used.
    version: The version set the number of instances for, if `None` the current
      version will be used.

  Raises:
    `InvalidVersionError` if the given module version isn't valid,
    `TransientError` if there is an issue persisting the change.
    `TypeError` if the given instances type is invalid.
  """
  rpc = set_num_instances_async(instances, module, version)
  rpc.get_result()


def set_num_instances_async(
    instances,
    module=None,
    version=None):
  """Returns a `UserRPC` to set the number of instances on the module version.

  Args:
    instances: The number of instances to set.
    module: The module to set the number of instances for, if `None` the current
      module will be used.
    version: The version set the number of instances for, if `None` the current
      version will be used.

  Returns:
    A `UserRPC` to set the number of instances on the module version.
  """

  def _ResultHook(rpc):
    mapped_errors = [
        modules_service_pb2.ModulesServiceError.INVALID_VERSION,
        modules_service_pb2.ModulesServiceError.TRANSIENT_ERROR
    ]
    _CheckAsyncResult(rpc, mapped_errors, {})

  if not isinstance(instances, six.integer_types):
    raise TypeError("'instances' arg must be of type long or int.")
  request = modules_service_pb2.SetNumInstancesRequest()
  request.instances = instances
  if module:
    request.module = module
  if version:
    request.version = version
  response = modules_service_pb2.SetNumInstancesResponse()
  return _MakeAsyncCall('SetNumInstances', request, response, _ResultHook)


def start_version(module, version):
  """Start all instances for the given version of the module.

  Args:
    module: String containing the name of the module to affect.
    version: String containing the name of the version of the module to start.

  Raises:
    `InvalidVersionError` if the given module version is invalid.
    `TransientError` if there is a problem persisting the change.
  """
  rpc = start_version_async(module, version)
  rpc.get_result()


def start_version_async(
    module,
    version):
  """Returns a `UserRPC` to start all instances for the given module version.

  Args:
    module: String containing the name of the module to affect.
    version: String containing the name of the version of the module to start.

  Returns:
    A `UserRPC` to start all instances for the given module version.
  """

  def _ResultHook(rpc):
    mapped_errors = [
        modules_service_pb2.ModulesServiceError.INVALID_VERSION,
        modules_service_pb2.ModulesServiceError.TRANSIENT_ERROR
    ]
    expected_errors = {
        modules_service_pb2.ModulesServiceError.UNEXPECTED_STATE:
            'The specified module: %s, version: %s is already started.' %
            (module, version)
    }
    _CheckAsyncResult(rpc, mapped_errors, expected_errors)

  request = modules_service_pb2.StartModuleRequest()
  request.module = module
  request.version = version
  response = modules_service_pb2.StartModuleResponse()
  return _MakeAsyncCall('StartModule', request, response, _ResultHook)


def stop_version(
    module=None,
    version=None):
  """Stops all instances for the given version of the module.

  Args:
    module: The module to affect, if `None` the current module is used.
    version: The version of the given module to affect, if `None` the current
      version is used.

  Raises:
    `InvalidVersionError` if the given module version is invalid.
    `TransientError` if there is a problem persisting the change.
  """
  rpc = stop_version_async(module, version)
  rpc.get_result()


def stop_version_async(
    module=None,
    version=None):
  """Returns a `UserRPC` to stop all instances for the given module version.

  Args:
    module: The module to affect, if `None` the current module is used.
    version: The version of the given module to affect, if `None` the current
      version is used.

  Returns:
    A `UserRPC` to stop all instances for the given module version.
  """

  def _ResultHook(rpc):
    mapped_errors = [
        modules_service_pb2.ModulesServiceError.INVALID_VERSION,
        modules_service_pb2.ModulesServiceError.TRANSIENT_ERROR
    ]
    expected_errors = {
        modules_service_pb2.ModulesServiceError.UNEXPECTED_STATE:
            'The specified module: %s, version: %s is already stopped.' %
            (module, version)
    }
    _CheckAsyncResult(rpc, mapped_errors, expected_errors)

  request = modules_service_pb2.StopModuleRequest()
  if module:
    request.module = module
  if version:
    request.version = version
  response = modules_service_pb2.StopModuleResponse()
  return _MakeAsyncCall('StopModule', request, response, _ResultHook)


def get_hostname(
    module=None,
    version=None,
    instance=None):
  """Returns a hostname to use to contact the module.

  Args:
    module: Name of module, if None, take module of the current instance.
    version: Name of version, if version is `None` then either use the version
      of the current instance if that version exists for the target module,
      otherwise use the default version of the target module.
    instance: Instance to construct a hostname for, if instance is None, a
      loadbalanced hostname for the module will be returned.  If the target
      module is not a fixed module, then instance is not considered valid.

  Returns:
    A valid canonical hostname that can be used to communicate with the given
    module/version/instance.  For example: `0.v1.module5.myapp.appspot.com`

  Raises:
    InvalidModuleError: if the given module version is invalid.
    InvalidInstancesError: if the given instance value is invalid.
    TypeError: if the given instance type is invalid.
  """

  def _ResultHook(rpc):
    mapped_errors = [
        modules_service_pb2.ModulesServiceError.INVALID_MODULE,
        modules_service_pb2.ModulesServiceError.INVALID_INSTANCES
    ]
    _CheckAsyncResult(rpc, mapped_errors, [])
    return rpc.response.hostname

  request = modules_service_pb2.GetHostnameRequest()
  if module:
    request.module = module
  if version:
    request.version = version
  if instance or instance == 0:
    if not isinstance(instance, (six.string_types, six.integer_types)):
      raise TypeError("'instance' arg must be of type basestring, long or int.")
    request.instance = str(instance)
  response = modules_service_pb2.GetHostnameResponse()
  return _MakeAsyncCall('GetHostname', request, response,
                        _ResultHook).get_result()
